/*--------------------------------------------------------------------------
----------------------------------------------------------------------------
--                                                                        --
--     Program Title : BigInt Version 0.4                                 --
--     Author : nex                                                       --
--     Date Last Modified : 1st January 2005                              --
--                                                                        --
--     Description : This program is for the handling of large integers   --
--                   by using a linked list representation. The intended  --
--                   purpose is to provent overflow when wanting to deal  --
--                   with very large integers.                            --
--                                                                        --
--     State of Program: The program is still being tested and has only   --
--                       the functions I have thus far required. It does  --
--                       not contain all the functions one might expect   --
--                       from such a package.                             --
--                                                                        --
----------------------------------------------------------------------------
--------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include "bigint.h"


/**************************************************************************/
/*---- Function defintions -----------------------------------------------*/

  /* Definitions for the use of the representation of big integers */


    /************************************************************************/
    /* Make a new 'big' integer type with value 'val' (which must be less 
       than 1000 */
    struct big MakeLittleBig(int val);


    /************************************************************************/
    /* Outputs the value of the 'big' type 'x' to standard output. */
    void WriteBig(struct big x);


    /************************************************************************/
    /* Returns the boolean expression x=y */
    int IsEqual(struct big x, struct big y);


    /************************************************************************/
    /* Calculates x + y (both of type 'big') and returns as type 'big'. This 
       function is specifically designed for the big type. */
    struct big Add(struct big x, struct big y);


    /************************************************************************/
    /* This is required to create new 'nodes' in the linked list (i.e. it
       allocates space in memory for next node) */
   struct big_list *lalloc(void);


/**************************************************************************/



/**************************************************************************/
/*---- Big integer specific functions ------------------------------------*/


  /************************************************************************/
  /* Make a new 'big' integer type with value 'val' (which must be less 
     than 1000 */

  struct big MakeLittleBig(int val)
    {
      struct big_list *One_L;
      struct big One;

      One_L = NULL;
      One_L = lalloc();
      One_L->Val = val;
      One_L->Next = One_L->Prev = NULL;
      One.Head = One.Tail = One_L;
      return One;
    }

  /************************************************************************/



  /************************************************************************/
  /* Outputs the value of the 'big' type 'x' to standard output. */

  void WriteBig(struct big x)
    {
      struct big X_Temp;

      X_Temp = x;
      if (x.Tail != NULL) {

	/* Extra zeros will need to be printed if current node value is less
           than 100 */
        if (x.Tail->Next != NULL) {
          if (x.Tail->Val < 100) {
	    if (x.Tail->Val < 10)
	      printf("00");
            else
	      printf("0");
          }
        }
	/* Output current node value then recursively call other node 
           values for output. */
        printf("%d", (x.Tail->Val));
        X_Temp.Tail = x.Tail->Prev;
        WriteBig(X_Temp); 
      }
    }

  /************************************************************************/



  /************************************************************************/
  /* Returns the boolean expression x=y */

  int IsEqual(struct big x, struct big y)
    {
      struct big X_Temp, Y_Temp;

      X_Temp = x;
      Y_Temp = y;

      /* Test all possible cases in which x == y (or x != y),
         using recursion to call all nodes. */
      if((X_Temp.Head != NULL) && (Y_Temp.Head != NULL)) {
        if (X_Temp.Head->Val == Y_Temp.Head->Val) {
          X_Temp.Head = X_Temp.Head->Next;
          Y_Temp.Head = Y_Temp.Head->Next;
          return (IsEqual(X_Temp, Y_Temp));
        } 
        else
          return 0;
      }
      else if ((X_Temp.Head != NULL) && (Y_Temp.Head == NULL))
        return 0;
      else if ((X_Temp.Head == NULL) && (Y_Temp.Head != NULL))
        return 0;
      else if ((X_Temp.Head == NULL) && (Y_Temp.Head == NULL))
        return 1;
      else
        return 0;
    }

  /************************************************************************/



  /************************************************************************/
  /* Calculates x + y (both of type 'big') and returns as type 'big'. This 
     function is specifically designed for the big type. */

  struct big Add(struct big x, struct big y)
    {
      /* Clear variables of data */
      struct big X_Temp, Y_Temp, R, R_Temp;
      int C, S, P, Q;

      C = S = P = Q = 0;
      R_Temp.Head = R_Temp.Tail = R.Head = R.Tail = NULL;

      X_Temp = x;
      Y_Temp = y;

      /* If x or y have no values, return R (which will == null) */
      if ((x.Head != NULL) || (y.Head != NULL)) {

	/* If there is still data in the current X_Temp or Y_Temp or there
           is a carry, get intermediate values for use in equations */
        while((X_Temp.Head != NULL) || (Y_Temp.Head != NULL) || (C == 1)) {
          if ((X_Temp.Head == NULL) && (Y_Temp.Head != NULL)) {
	    Q = Y_Temp.Head->Val;
	    Y_Temp.Head = Y_Temp.Head->Next;
	    P = 0;
          }
          else if ((X_Temp.Head != NULL) && (Y_Temp.Head == NULL)) {
	    P = X_Temp.Head->Val;
            X_Temp.Head = X_Temp.Head->Next;
            Q = 0;
          }
          else if ((X_Temp.Head == NULL) && (Y_Temp.Head == NULL)) {
            P = Q = 0;
          }
          else {
            P = X_Temp.Head->Val;
            X_Temp.Head = X_Temp.Head->Next;
            Q = Y_Temp.Head->Val;
            Y_Temp.Head = Y_Temp.Head->Next;
          }
      
	  /* Use intermediate values to calculate sum and carry */
          if (R.Head == NULL) {
	    S = ((P + Q) % 1000);
	    C = ((P + Q) / 1000);
          }
          else {
	    S = ((P + Q + C) % 1000);
	    C = ((P + Q + C) / 1000);
          }
      
	  /* Add new nodes containing correct values to doubly linked 
             list */
          if (R.Head == NULL) {
	    R.Head = lalloc();
	    R.Head->Val = S;
	    R.Head->Next = R.Head->Prev = NULL;
	    R.Tail = R.Head;
          }
          else {
	    R.Tail->Next = lalloc();
	    R.Tail->Next->Val = S;
	    R.Tail->Next->Next = R.Tail->Next->Prev = NULL;
	    R_Temp = R;
	    R.Tail = R.Tail->Next;
	    R.Tail->Prev = R_Temp.Tail;
          }
        }
      }
      return R;
    }

  /************************************************************************/



  /************************************************************************/
  /* This is required to create new 'nodes' in the linked list (i.e. it
     allocates space in memory for next node) */

  struct big_list *lalloc(void)
    {
      return (struct big_list *) malloc(sizeof(struct big_list));
    }

  /************************************************************************/



/*---- End of Big integer specific functions -----------------------------*/
/**************************************************************************/
