/*--------------------------------------------------------------------------
----------------------------------------------------------------------------
--                                                                        --
--     Program Title : Fast-Fib Version 2.4                               --
--     Author : nex                                                       --
--     Date Last Modified : 1st January 2005                              --
--                                                                        --
--     Description : This program takes in two integers from standard     --
--                   input and uses them as indices to calculate          --
--                   Fibonacci numbers from the lower to upper index.     --
--                                                                        --
--                   This program uses a large integer representation     --
--                   for large integer values produced by the Fibonacci   --
--                   algorithm (to prevent overflow) contained in         --
--                   bigint.h.                                            --
--                                                                        --
--     State of Program: The program has worked correctly for all test    --
--                       data.                                            --
--                                                                        --
----------------------------------------------------------------------------
--------------------------------------------------------------------------*/

#include <stdio.h>
#include "bigint.h"




/**************************************************************************/
/*---- Function defintions -----------------------------------------------*/

  /* Definitions for the fast Fibonacci algorithm functions */
    struct big helpFib(int n, int k, struct big fibk, struct big fibk1);
    struct big Fib(int n);
    int Verify(int n);

/**************************************************************************/






/**************************************************************************/
/**************************************************************************/
int main(void)
  {  

    int LowFib, HiFib, Swap, i;

    /* Get Lower and Upper values to run algorithm from and to */
    printf("\nEnter low number: ");
    scanf("%d", &LowFib);
    printf("Enter high number: ");
    scanf("%d", &HiFib);
    printf("\n");

    /* Check that Lower and Upper values are positive */
    if ((LowFib < 1) || (HiFib < 1)) {
      printf("\nFibonacci Indices must be positive integers greater than"); 
      printf(" zero.");
    }
    else {
      /* Check that Lower Limit is less than the Upper Limit and swap if */
      /* appropriate */
      if (LowFib > HiFib) {
	Swap = LowFib;
	LowFib = HiFib;
	HiFib = Swap;
      }

      /* Find, validate and output requested Fibonacci sequence */
      for ( i = LowFib; i <= HiFib; i++) {
	printf("Fib(%d) = ", i);
	WriteBig(Fib(i));
	if (Verify(i))
	  printf("    Valid\n");
	else
	  printf("    Invalid\n");
      }
    }

    /* Prevent unwanted output messages at end of program */
    return 0;
  }

/**************************************************************************/
/**************************************************************************/






/**************************************************************************/
/*---- Fibonacci specific functions --------------------------------------*/

  /************************************************************************/
  /* helpFib is used to speed up the Fib function by preventing repeated 
     calculations. In truth, I only ever briefly understood how this works. 
     I didn't come up with it and I understood it just long enough to 
     program it in the original Ada version of this program. */

  struct big helpFib(int n, int k, struct big fibk, struct big fibk1)
    {
      if (n == k)
        return fibk;
      else
        return (helpFib(n,(k + 1), Add(fibk, fibk1), fibk));
    }

  /************************************************************************/



  /************************************************************************/
  /* The Fib function calculates the value of the Fibonacci number with 
     index n. */

  struct big Fib(int n)
    {
      struct big Big_One, Big_Zero;

      Big_One = MakeLittleBig(1);
      Big_Zero = MakeLittleBig(0);

      /* For values less than 3, the Fib number is 1 */
      if (n < 3)
        return Big_One;

      /* For Fibonacci indices smaller than 21, use a simple recursive Fib 
         function as this produces the most optimal peformance */
      else if (n < 21)
        return (Add((Fib(n-2)), (Fib(n-1))));

      /* For indices larger than 20, use the helpFib function to
         speed things up. */
      else
        return (helpFib(n, 1, Big_One, Big_Zero));
    }

  /************************************************************************/



  /************************************************************************/
  /* The Verify procedure tests to see if the Fibonacci number
     calculated is accurate by trying to find it by a different 
     algorithm (which, incidentally, uses the same algorithm as the 'main
     program' as it were) and then comparing it to the originally found 
     value. */

  int Verify(int n)
    {
      struct big test;

      if (n != 2) {
      test = Fib(n);

      /* The 'Add' function is required to add numbers represented using
         the 'big' structure */
      test = Add(Add(test,test),test);

      /* The 'IsEqual' function is required to compare numbers represented 
         using the 'big' structure */
      if (IsEqual(test, Add(Fib(n+2), Fib(n-2))))
        return 1;
      else
        return 0;}
      else
        return 1;
    }

/**************************************************************************/

/*---- End of Fibonacci specific functions -------------------------------*/
/**************************************************************************/


/*---- End of Fast-Fib program -------------------------------------------*/
/**************************************************************************/
